#!/usr/bin/env python3
# vim: foldmethod=marker

import copy


# Node {{{1
class Node:
    def __init__(self, key, *, data=None, next=None):
        self.key = key
        self.data = data
        self.next = next


# LinkedList {{{1
class LinkedList:
    # List creation {{{2
    def __init__(self, key, data, *, reverse=False):
        start = Node(key[0], data=data[0])  # create root node
        end = start
        for k, d in zip(key[1:], data[1:]):
            if reverse:
                end = self._prepend(end, Node(k, data=d))
            else:
                end = self._append(end, Node(k, data=d))
        self.first = end if reverse else start
        self.last = start if reverse else end

    @staticmethod
    def _append(anchor, node):
        anchor.next = node
        return node

    @staticmethod
    def _prepend(anchor, node):
        node.next = anchor
        return node

    # Sequence properties {{{2
    def __getitem__(self, key):
        _, node = self._getitem(key)
        return node

    def __len__(self):
        node = self.first
        count = 0
        while node != None:
            count += 1
            node = node.next
        return count

    def __str__(self):
        node = self.first
        pretty = ""
        while node != None:
            pretty += f"node: {hex(id(node))} : key={node.key} : data={node.data}\n"
            node = node.next
        return pretty

    def _getitem(self, key):
        """Private method to retrieve a list element given a key.
           Arguments:
              key: identifies the node
           Returns:
              Tuple with reference to predecessor node and reference to node
              with given key
        """
        node = self.first
        prev = node
        while node != None:
            if node.key == key:
                return (prev, node)
            prev = node
            node = node.next
        raise KeyError(f"Key '{key}' not found")

    # Node insertion {{{2
    def insert(self, key, node, *, before=False):
        anchor = self.__getitem__(key)
        if before:  # insert before
            tmp = copy.copy(anchor)  # try to understand what happens if you do not create a (shallow) copy
            anchor.key = node.key
            anchor.data = node.data
            node = tmp # overwrite reference of node passed as argument (garbage collected by python)
            anchor.next = node
        else:  # insert after
            assert anchor is not None
            node.next = anchor.next
            anchor.next = node
        return node

    # Node removal {{{2
    def remove(self, key):
        if self.first.key == key:
            del_node = self.first
            self.first = self.first.next
        else:
            prev, del_node = self._getitem(key)
            prev.next = del_node.next
        del del_node

    # Iterator {{{2
    def __iter__(self):
        return LinkedListForwardIterator(self.first)


# LinkedListForwardIterator {{{1
class LinkedListForwardIterator:
    def __init__(self, start):
        self.node = start

    def __next__(self):
        if self.node != None:
            curr_node = self.node
            self.node = self.node.next
            return curr_node
        else:
            raise StopIteration

    def __iter__(self):
        return self


# main() {{{1
def main():
    # Create some dummy data
    class Data:
        def __init__(self, data):
            self.data = data

        def __str__(self):
            return f'<value={self.data} {hex(id(self))}>'

    data = [Data(d) for d in list('ABCDE')]
    keys = [f'ID_{k}' for k in range(len(data))]

    # create a list
    ll = LinkedList(keys, data)
    print('Initial linked list')
    print(len(ll))
    print(ll)

    # insert a new node
    ll.insert(keys[1], Node(10, data='Data can be arbitrary'))
    print('After insertion of new node with key=10')
    print(len(ll))
    print(ll)

    # remove a node
    ll.remove(10)
    print('After removal of node with key=10 (same as initial list)')
    print(len(ll))
    print(ll)

    # Sequence protocol satisfied:
    print('LinkedListForwardIterator usage in for-loop:')
    for node in ll:
        print(node)


if __name__ == "__main__":
    main()
